from predicthq.endpoints.base import BaseEndpoint
from .schemas import (
    CreateAnalysisResponse,
    AnalysisResultSet,
    Analysis,
    FeatureImportance,
    CorrelationResultSet,
    CreateAnalysisGroupResponse,
    AnalysisGroup,
    AnalysisGroupResultSet,
)
from predicthq.endpoints.decorators import accepts, returns
from typing import overload, List, Optional, TextIO, Union
from datetime import date


class BeamEndpoint:
    def __init__(self, client):
        self.analysis = self.AnalysisEndpoint(client)
        self.analysis_group = self.AnalysisGroupEndpoint(client)

    class AnalysisEndpoint(BaseEndpoint):
        @overload
        def create(
            self,
            name: str,
            location__geopoint: dict,
            location__radius: Optional[float] = None,
            location__unit: Optional[str] = None,
            location__google_place_id: Optional[str] = None,
            location__geoscope_paths: Optional[List[str]] = None,
            rank__type: Optional[str] = None,
            rank__levels__phq: Optional[dict] = None,
            rank__levels__local: Optional[dict] = None,
            demand_type__industry: Optional[str] = None,
            demand_type__unit_descriptor: Optional[str] = None,
            demand_type__unit_currency_multiplier: Optional[float] = None,
            demand_type__currency_code: Optional[str] = None,
            tz: Optional[str] = None,
            external_id: Optional[str] = None,
            label: Optional[List[str]] = None,
            **params,
        ): ...
        @accepts(query_string=False)
        @returns(CreateAnalysisResponse)
        def create(self, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.post(
                f"{self.build_url('v1', 'beam')}analyses/",
                json=params,
                verify=verify_ssl,
            )

        @overload
        def search(
            self,
            updated__gt: Optional[str] = None,
            updated__gte: Optional[str] = None,
            updated__lt: Optional[str] = None,
            updated__lte: Optional[str] = None,
            q: Optional[str] = None,
            status: Optional[List[str]] = None,
            group_id: Optional[List[str]] = None,
            demand_type__interval: Optional[List[str]] = None,
            demand_type__industry: Optional[List[str]] = None,
            readiness_status: Optional[List[str]] = None,
            include_deleted: Optional[bool] = None,
            sort: Optional[List[str]] = None,
            offset: Optional[int] = None,
            limit: Optional[int] = None,
            external_id: Optional[List[str]] = None,
            label: Optional[List[str]] = None,
            **params,
        ): ...
        @accepts()
        @returns(AnalysisResultSet)
        def search(self, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.get(
                f"{self.build_url('v1', 'beam')}analyses/",
                params=params,
                verify=verify_ssl,
            )

        @accepts()
        @returns(Analysis)
        def get(self, analysis_id, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.get(
                f"{self.build_url('v1', 'beam')}analyses/{analysis_id}/",
                params=params,
                verify=verify_ssl,
            )

        @overload
        def update(
            self,
            analysis_id: str,
            name: Optional[str] = None,
            location__geopoint: Optional[dict] = None,
            location__radius: Optional[float] = None,
            location__unit: Optional[str] = None,
            location__google_place_id: Optional[str] = None,
            location__geoscope_paths: Optional[List[str]] = None,
            rank__type: Optional[str] = None,
            rank__levels__phq: Optional[dict] = None,
            rank__levels__local: Optional[dict] = None,
            demand_type__industry: Optional[str] = None,
            demand_type__unit_descriptor: Optional[str] = None,
            demand_type__unit_currency_multiplier: Optional[float] = None,
            demand_type__currency_code: Optional[str] = None,
            tz: Optional[str] = None,
            external_id: Optional[str] = None,
            label: Optional[List[str]] = None,
            **params,
        ): ...
        @accepts(query_string=False)
        def update(self, analysis_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.patch(
                f"{self.build_url('v1', 'beam')}analyses/{analysis_id}/",
                json=params,
                verify=verify_ssl,
            )

        @accepts()
        def delete(self, analysis_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.delete(
                f"{self.build_url('v1', 'beam')}analyses/{analysis_id}/",
                params=params,
                verify=verify_ssl,
            )

        @accepts()
        def refresh(self, analysis_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.post(
                f"{self.build_url('v1', 'beam')}analyses/{analysis_id}/refresh/",
                params=params,
                verify=verify_ssl,
            )

        @overload
        def get_correlation_results(
            self,
            analysis_id: str,
            date__gt: Optional[date] = None,
            date__gte: Optional[date] = None,
            date__lt: Optional[date] = None,
            date__lte: Optional[date] = None,
            offset: Optional[int] = None,
            limit: Optional[int] = None,
            include_features: Optional[List[str]] = None,
            **params,
        ): ...
        @accepts()
        @returns(CorrelationResultSet)
        def get_correlation_results(self, analysis_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.get(
                f"{self.build_url('v1', 'beam')}analyses/{analysis_id}/correlate/",
                params=params,
                verify=verify_ssl,
            )

        @accepts()
        @returns(FeatureImportance)
        def get_feature_importance(self, analysis_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.get(
                f"{self.build_url('v1', 'beam')}analyses/{analysis_id}/feature-importance/",
                params=params,
                verify=verify_ssl,
            )

        @accepts(query_string=False)
        @overload
        def upload_demand(
            self,
            analysis_id: str,
            json: Optional[Union[str, TextIO]] = None,
            ndjson:  Optional[Union[str, TextIO]] = None,
            csv:  Optional[Union[str, TextIO]] = None,
            **params,
        ): ...
        def upload_demand(self, analysis_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            if json := params.pop("json", None):
                headers = {"Content-Type": "application/json"}
                file = json
            if ndjson := params.pop("ndjson", None):
                headers = {"Content-Type": "application/x-ndjson"}
                file = ndjson
            if csv := params.pop("csv", None):
                headers = {"Content-Type": "text/csv"}
                file = csv

            if isinstance(file, str):
                with open(file) as f:
                    return self.client.post(
                        f"{self.build_url('v1', 'beam')}analyses/{analysis_id}/sink/",
                        data=f.read(),
                        headers=headers,
                        verify=verify_ssl,
                    )

            return self.client.post(
                f"{self.build_url('v1', 'beam')}analyses/{analysis_id}/sink/",
                data=file.read(),
                headers=headers,
                verify=verify_ssl,
            )

    class AnalysisGroupEndpoint(BaseEndpoint):
        @overload
        def create(
            self,
            name: str,
            analysis_ids: List[str],
            demand_type__unit_descriptor: Optional[str] = None,
            **params,
        ): ...
        @accepts(query_string=False)
        @returns(CreateAnalysisGroupResponse)
        def create(self, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.post(
                f"{self.build_url('v1', 'beam')}analysis-groups/",
                json=params,
                verify=verify_ssl,
            )

        @overload
        def search(
            self,
            updated__gt: Optional[str] = None,
            updated__gte: Optional[str] = None,
            updated__lt: Optional[str] = None,
            updated__lte: Optional[str] = None,
            q: Optional[str] = None,
            status: Optional[List[str]] = None,
            demand_type__interval: Optional[List[str]] = None,
            demand_type__industry: Optional[List[str]] = None,
            readiness_status: Optional[List[str]] = None,
            include_deleted: Optional[bool] = None,
            sort: Optional[List[str]] = None,
            offset: Optional[int] = None,
            limit: Optional[int] = None,
            **params,
        ): ...
        @accepts()
        @returns(AnalysisGroupResultSet)
        def search(self, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.get(
                f"{self.build_url('v1', 'beam')}analysis-groups/",
                params=params,
                verify=verify_ssl,
            )

        @accepts()
        @returns(AnalysisGroup)
        def get(self, group_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.get(
                f"{self.build_url('v1', 'beam')}analysis-groups/{group_id}/",
                params=params,
                verify=verify_ssl,
            )

        @overload
        def update(
            self,
            group_id: str,
            name: Optional[str] = None,
            analysis_ids: Optional[List[str]] = None,
            demand_type__unit_descriptor: Optional[str] = None,
            **params,
        ): ...
        @accepts(query_string=False)
        def update(self, group_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.patch(
                f"{self.build_url('v1', 'beam')}analysis-groups/{group_id}/",
                json=params,
                verify=verify_ssl,
            )

        @accepts()
        def delete(self, group_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.delete(
                f"{self.build_url('v1', 'beam')}analysis-groups/{group_id}/",
                params=params,
                verify=verify_ssl,
            )

        @accepts()
        def refresh(self, group_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.post(
                f"{self.build_url('v1', 'beam')}analysis-groups/{group_id}/refresh/",
                params=params,
                verify=verify_ssl,
            )

        @accepts()
        @returns(FeatureImportance)
        def get_feature_importance(self, group_id: str, **params):
            verify_ssl = params.pop("config.verify_ssl", True)
            return self.client.get(
                f"{self.build_url('v1', 'beam')}analysis-groups/{group_id}/feature-importance/",
                params=params,
                verify=verify_ssl,
            )



from predicthq.endpoints.base import UserBaseEndpoint
from predicthq.endpoints.decorators import accepts, returns

from .schemas import FeatureResultSet


class FeaturesEndpoint(UserBaseEndpoint):
    BASE_FEATURE_CRITERIA = {"stats": [["sum", "count"]], "phq_rank": [None]}
    FIELDS_TO_MUTATE = frozenset(
        [
            "phq_attendance_",
            "phq_viewership_",
            "phq_impact_",
            "phq_spend_",
        ]
    )

    @classmethod
    def mutate_bool_to_default_for_type(cls, user_request_spec):
        for key, val in user_request_spec.items():
            if any(key.startswith(x) for x in cls.FIELDS_TO_MUTATE):
                user_request_spec[key] = [
                    cls.BASE_FEATURE_CRITERIA if isinstance(v, bool) else v for v in val
                ]

    @accepts(query_string=False)
    @returns(FeatureResultSet)
    # This is a temporary solution to get the next page for Features API
    # _params and _json are for internal use only
    def obtain_features(self, _params: dict = None, _json: dict = None, **request):
        verify_ssl = request.pop("config", {}).get("verify_ssl", True)
        return self.client.post(
            self.build_url("v1", "features"),
            json=_json or request,
            verify=verify_ssl,
            params=_params,
        )



import functools

from pydantic import ValidationError as PydanticValidationError

from predicthq.exceptions import ValidationError

from predicthq.endpoints.schemas import ArgKwargResultSet


def _kwargs_to_key_list_mapping(kwargs, separator="__"):
    """
    Converts kwargs to a nested dictionary mapping keys to lists of values
    """
    data = {}
    for key, value in kwargs.items():
        keys = key.split(separator, 1)
        if len(keys) > 1:
            value = {keys[1]: value}
        if isinstance(value, dict):
            value = _kwargs_to_key_list_mapping(value)

        data[keys[0]] = [] if not data.get(keys[0]) else data[keys[0]]
        data[keys[0]].append(value)
    return data


def _to_url_params(key_list_mapping, glue=".", separator=",", parent_key=""):
    """
    Converts key_list_mapping to url parameters
    """
    params = {}
    for key, value in key_list_mapping.items():
        current_key = f"{parent_key}{glue}{key}" if parent_key else key
        for v in value:
            if isinstance(v, dict):
                params.update(_to_url_params(v, glue, separator, current_key))
            elif isinstance(v, list):
                params.update({current_key: separator.join(map(str, v))})
            elif isinstance(v, bool):
                params.update({current_key: 1 if v else 0})
            else:
                params.update({current_key: v})
    return params


def _to_json(key_list_mapping, json=None):
    """
    Converts key_list_mapping to json
    """
    if json is None:
        json = {}
    for key, value in key_list_mapping.items():
        for v in value:
            json[key] = dict() if not json.get(key) else json[key]
            if isinstance(v, dict):
                _to_json(v, json[key])
            else:
                json[key] = v
    return json


def accepts(query_string=True, role=None):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(endpoint, *args, **kwargs):
            key_list_mapping = _kwargs_to_key_list_mapping(kwargs)
            if hasattr(endpoint, "mutate_bool_to_default_for_type"):
                endpoint.mutate_bool_to_default_for_type(key_list_mapping)

            if query_string:
                data = _to_url_params(key_list_mapping)
            else:
                data = _to_json(key_list_mapping)

            return f(endpoint, *args, **data)

        return wrapper

    return decorator


def returns(model_class):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(endpoint, *args, **kwargs):
            model = getattr(endpoint.Meta, f.__name__, {}).get("returns", model_class)

            data = f(endpoint, *args, **kwargs)
            try:
                loaded_model = model(**data)
                loaded_model._more = functools.partial(wrapper, endpoint)
                loaded_model._endpoint = endpoint
                if isinstance(loaded_model, ArgKwargResultSet):
                    loaded_model._args = args
                    loaded_model._kwargs = kwargs
                return loaded_model
            except PydanticValidationError as e:
                raise ValidationError(e)

        return wrapper

    return decorator



from pydantic import BaseModel, Field, ConfigDict
from datetime import datetime
from predicthq.endpoints.schemas import ArgKwargResultSet
from typing import Optional, List


class BeamPaginationResultSet(ArgKwargResultSet):
    def has_next(self):
        return self._kwargs.get("offset", 0) + len(self.results) < self.count

    def get_next(self):
        if "offset" in self._kwargs:
            self._kwargs["offset"] = self._kwargs.get("offset") + len(self.results)
        else:
            self._kwargs["offset"] = len(self.results)
        return self._more(**self._kwargs)



class CreateAnalysisResponse(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    analysis_id: str


class GeoPoint(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    lat: str
    lon: str


class Location(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    geopoint: GeoPoint
    radius: float
    unit: str
    google_place_id: Optional[str] = None


class RankLevel(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    min: int
    max: Optional[int] = None


class RankLevels(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    phq: Optional[RankLevel] = None
    local: Optional[RankLevel] = None


class Rank(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    type: str
    levels: Optional[RankLevels] = None


class AnalysisDateRange(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    start: datetime
    end: datetime


class BestPracticeChecks(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    industry: bool = False
    rank: bool = False
    radius: bool = False


class AnalysisReadinessChecks(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    date_range: Optional[AnalysisDateRange] = None
    error_code: Optional[str] = None
    missing_dates: Optional[List[str]] = None
    validation_response: Optional[dict] = None
    best_practice: bool
    best_practice_checks: BestPracticeChecks


class ProcessingCompleted(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    correlation: bool
    feature_importance: bool
    value_quant: bool


class DemandTypeGroup(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    interval: str
    week_start_day: Optional[str] = None
    industry: Optional[str] = None
    unit_descriptor: Optional[str] = None
    currency_code: str


class DemandType(DemandTypeGroup):
    interval: str
    week_start_day: Optional[str] = None
    industry: Optional[str] = None
    unit_descriptor: Optional[str] = None
    unit_currency_multiplier: float
    currency_code: str


class Analysis(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    analysis_id: Optional[str] = None
    name: str
    location: Location
    rank: Rank
    user_id: Optional[str] = None
    access_type: str
    status: str
    readiness_status: Optional[str] = None
    readiness_checks: AnalysisReadinessChecks
    processing_completed: ProcessingCompleted
    demand_type: DemandType
    group_ids: Optional[List[str]] = None
    tz: Optional[str] = None
    create_dt: datetime
    update_dt: datetime
    processed_dt: Optional[datetime] = None
    external_id: Optional[str] = None
    label: Optional[List[str]] = None


class AnalysisResultSet(BeamPaginationResultSet):
    results: List[Analysis] = Field(alias="analyses")


class FeatureGroup(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    feature_group: str
    features: List[str]
    p_value: float
    important: bool


class FeatureImportance(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    feature_importance: List[FeatureGroup]


class CorrelationResult(BaseModel):
    date: str
    actual_demand: Optional[float] = None
    baseline_demand: Optional[float] = None
    remainder: Optional[float] = None
    impact_significance: str
    impact_significance_score: Optional[int] = None


class CorrelationResultSet(BeamPaginationResultSet):
    model_config = ConfigDict(protected_namespaces=())
    model_version: str
    version: int
    results: List[CorrelationResult] = Field(alias="dates")


class CreateAnalysisGroupResponse(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    group_id: str


class ExcludedAnalysis(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    analysis_id: str
    reason: str
    excluded_from: List[str]


class ProcessingCompletedGroup(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    feature_importance: bool
    value_quant: bool
    excluded_analyses: List[ExcludedAnalysis]


class AnalysisGroup(BaseModel):
    model_config: ConfigDict = ConfigDict(extra="allow")

    group_id: Optional[str] = None
    name: str
    analysis_ids: List[str]
    user_id: Optional[str] = None
    processing_completed: ProcessingCompletedGroup
    readiness_status: Optional[str] = None
    demand_type: Optional[DemandTypeGroup] = None
    status: str
    create_dt: datetime
    update_dt: datetime
    processed_dt: Optional[datetime] = None


class AnalysisGroupResultSet(BeamPaginationResultSet):
    results: List[AnalysisGroup] = Field(alias="groups")



import unittest
from unittest import mock

import pytest
import requests
import stamina

import predicthq
from predicthq import endpoints
from predicthq.endpoints.oauth2.schemas import AccessToken
from predicthq.exceptions import ClientError, RetriableError, ServerError
from tests import load_fixture, with_config, with_mock_client, with_mock_responses


class ClientTest(unittest.TestCase):
    def setUp(self):
        stamina.set_testing(True)
        self.client = predicthq.Client()

    def tearDown(self):
        stamina.set_testing(False)

    @with_config(ENDPOINT_URL="https://api.predicthq.com")
    def test_build_url(self):
        assert self.client.build_url("v1") == "https://api.predicthq.com/v1"

    def test_endpoints_initialization(self):
        assert isinstance(self.client.oauth2, endpoints.OAuth2Endpoint)
        assert isinstance(self.client.accounts, endpoints.AccountsEndpoint)
        assert isinstance(self.client.events, endpoints.EventsEndpoint)
        assert isinstance(self.client.places, endpoints.PlacesEndpoint)

    @with_mock_responses()
    def test_request(self, responses):
        self.client.logger.setLevel("DEBUG")

        assert self.client.get("/v1/") == responses.calls[0].response.json()

        assert self.client.get("/no-content/") is None

        assert self.client.get("/invalid-json/") is None

        with pytest.raises(ClientError) as excinfo:
            self.client.get("/not-found/")
        assert excinfo.value.args[0] == responses.calls[3].response.json()

        with pytest.raises(ServerError) as excinfo:
            self.client.get("/server-error/")
        assert excinfo.value.args[0] == responses.calls[4].response.json()

        with pytest.raises(ServerError) as excinfo:
            self.client.get("/no-json/")
        assert excinfo.value.args[0] == responses.calls[5].response.content

        # Test headers
        self.client.authenticate(client_id="client_id", client_secret="client_secret", scope=["account"])
        assert responses.calls[6].request.headers["Authorization"] == "Basic Y2xpZW50X2lkOmNsaWVudF9zZWNyZXQ="
        assert responses.calls[6].request.headers["Content-Type"] == "application/x-www-form-urlencoded"

        self.client.accounts.self()
        assert responses.calls[7].request.headers["Authorization"] == "Bearer token123"
        assert responses.calls[7].request.headers["Accept"] == "application/json"

    @with_mock_client(request_returns={"result": "value"})
    def test_get(self, client):
        result = self.client.get("/get/", params={"q": "query"})
        client.request.assert_called_once_with("get", "/get/", params={"q": "query"})
        assert result == client.request.return_value

    @with_mock_client(request_returns={"result": "value"})
    def test_post(self, client):
        result = self.client.post("/post/", data={"key": "value"})
        client.request.assert_called_once_with("post", "/post/", data={"key": "value"})
        assert result == client.request.return_value

    @with_mock_client(request_returns={"result": "value"})
    def test_put(self, client):
        result = self.client.put("/put/", data={"key": "value"})
        client.request.assert_called_once_with("put", "/put/", data={"key": "value"})
        assert result == client.request.return_value

    @with_mock_client(request_returns={"result": "value"})
    def test_patch(self, client):
        result = self.client.patch("/patch/", data={"key": "value"})
        client.request.assert_called_once_with("patch", "/patch/", data={"key": "value"})
        assert result == client.request.return_value

    @with_mock_client()
    def test_delete(self, client):
        result = client.delete("/delete/")
        client.request.assert_called_once_with("delete", "/delete/")
        assert result == client.request.return_value

    @with_mock_client(request_returns=load_fixture("access_token"))
    def test_authenticate(self, client):
        token = self.client.authenticate(
            client_id="client_id", client_secret="client_secret", scope=["account", "events"]
        )
        client.request.assert_called_once_with(
            "post",
            "/oauth2/token/",
            auth=("client_id", "client_secret"),
            data={"scope": "account events", "grant_type": "client_credentials"},
            verify=True,
        )

        assert isinstance(token, AccessToken)

        assert self.client.access_token == "token123"

    def test_construct_with_access_token(self):
        client = predicthq.Client(access_token="token123")
        assert client.access_token == "token123"

    @with_config(OAUTH2_ACCESS_TOKEN="token123")
    def test_construct_with_access_token_config(self):
        client = predicthq.Client()
        assert client.access_token == "token123"


# New tests are intentionally left out of the ClientTest class because pytest does not support
# parametrized fixtures inside unittest.TestCase classes as documented at
# https://docs.pytest.org/en/stable/how-to/unittest.html#pytest-features-in-unittest-testcase-subclasses
# We should gradually migrate all tests to pytest style and remove unittest.TestCase classes


@pytest.fixture
def client():
    return predicthq.Client()


@pytest.mark.parametrize("status_code", (429, 503, 504))
@mock.patch("predicthq.client.requests.request")
def test_retries(mock_request, client, status_code):
    stamina.set_testing(True, attempts=3)  # Disable stamina backoff waiting so the tests can run faster

    mock_request.return_value.status_code = status_code
    mock_request.return_value.raise_for_status.side_effect = requests.HTTPError

    with pytest.raises(RetriableError):
        client.get("/get/")

    # Check that the request was retried 3 times
    mock_request.assert_has_calls(
        [
            mock.call("get", client.build_url("/get/"), headers=client.get_headers({})),
            mock.call().raise_for_status(),
            mock.call().json(),
        ]
        * 3
    )

    stamina.set_testing(False)  # Enable stamina backoff waiting after the test is done



import csv
from datetime import date
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, RootModel

from predicthq.endpoints.schemas import ArgKwargResultSet


class CsvMixin:
    def _flatten_json(self, separator: str) -> Optional[List[dict]]:
        def __flatten_json(d: dict, pk: str = "") -> dict:
            flat_json = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    flat_json.update(
                        __flatten_json(v, f"{pk}{separator}{k}" if pk else k)
                    )
                    continue
                flat_json.update({f"{pk}{separator}{k}" if pk else k: v})
            return flat_json

        return [__flatten_json(d.model_dump(exclude_none=True)) for d in self.iter_all()]

    def to_csv(self, file: str, mode: str = "w+", separator: str = "_") -> None:
        header = None
        with open(file, mode=mode) as csv_file:
            csv_writer = csv.writer(csv_file)
            for d in self._flatten_json(separator):
                if not header:
                    header = list(d.keys())
                    csv_writer.writerow(header)
                csv_writer.writerow([d[h] for h in header])


class FeatureRankLevel(BaseModel):
    rank_levels: dict


class FeatureStats(BaseModel):
    avg: Optional[float] = None
    count: Optional[int] = None
    max: Optional[float] = None
    median: Optional[float] = None
    min: Optional[float] = None
    sum: Optional[float] = None
    std_dev: Optional[float] = None


class FeatureStat(BaseModel):
    stats: FeatureStats


class Feature(RootModel):
    root: Dict[str, Union[date, FeatureStat, FeatureRankLevel]]

    def __getattr__(self, name: str) -> Union[date, FeatureStat, FeatureRankLevel]:
        return self.root[name]


class FeatureResultSet(ArgKwargResultSet, CsvMixin):
    results: List[Optional[Feature]]

    def get_next(self):
        if not self.has_next() or not hasattr(self, "_more"):
            return
        params = self._parse_params(self.next)
        return self._more(_params=params, _json=self._kwargs.get("_json", {}) or self._kwargs)



from typing import List, Optional, Tuple

from pydantic import BaseModel, field_validator

from predicthq.endpoints.schemas import ResultSet


class Place(BaseModel):
    id: str
    type: str
    name: str
    county: Optional[str] = None
    region: Optional[str] = None
    country: Optional[str] = None
    country_alpha2: Optional[str] = None
    country_alpha3: Optional[str] = None
    location: Tuple[float, float]


class PlaceResultSet(ResultSet):
    results: List[Optional[Place]]



from .oauth2 import OAuth2Endpoint
from .v1.accounts import AccountsEndpoint
from .v1.broadcasts import BroadcastsEndpoint
from .v1.events import EventsEndpoint
from .v1.features import FeaturesEndpoint
from .v1.places import PlacesEndpoint
from .v1.radius import SuggestedRadiusEndpoint
from .v1.beam import BeamEndpoint


__all__ = [
    "OAuth2Endpoint",
    "AccountsEndpoint",
    "BroadcastsEndpoint",
    "EventsEndpoint",
    "FeaturesEndpoint",
    "PlacesEndpoint",
    "SuggestedRadiusEndpoint",
    "BeamEndpoint",
]
